#include <autoconf.h>
#include <hermit.h>
#include <herrno.h>
#include <init.h>
#include <platform.h>
#include <machine.h>
#include <arch/system.h>
#include <arch/pinmux.h>
#include <htypes.h>
#include <io.h>
#include <string.h>
#include <memdev.h>
#include <param.h>
#include <map.h>
#include <boost.h>
#include <flash.h>
#include <medium.h>
#include <led.h>
#include <i2c.h>
#include <i2c_rma1.h>
#include <net/eth_util.h>
#include <net/rma1.h>
#include <rma1_sdhc.h>
#include "board.h"

char target_name[256];
char *target_profile = target_name;

static struct board_private board_priv;
static struct memory_device mdev_param;

#define FLASH_ADDR(offset) (FLASH_START + (offset))
#define RAM_ADDR(offset) (DRAM_START + (offset))

#define xJP1	PIN_FCE0_N
#define xSW1	PIN_MEMC_AD6
#define xSW2	PIN_MEMC_AD7
#define xSW3	PIN_MEMC_AD4
#define xSW4	PIN_MEMC_AD5
#define xLED1	PIN_FRB
#define xLED2	PIN_CS4_N
#define xLED3	PIN_CS2_N
#define xLED4	PIN_WAIT_N

static void gpio_output(u32 pin, u8 level)
{
	if (level)
		write32(PIN_GPIODSR(pin), 1 << PIN_GPIO32(pin));
	else
		write32(PIN_GPIODCR(pin), 1 << PIN_GPIO32(pin));
}

static u8 gpio_input(u32 pin)
{
	u32 val;
	val = read32(PIN_GPIODCR(pin)) & (1 << PIN_GPIO32(pin));
	return val ? 1 : 0;
}

static char *get_boot_mode_str(u8 mode)
{
	switch (mode) {
	case BOOT_MODE_SD:	/* MD(2, 3, 4, 6, 7) = 0, 1, 0, 1, 0 */
		return "SD";
	case BOOT_MODE_EMMC:	/* MD(2, 3, 4, 6, 7) = 1, 1, 0, 1, 0 */
		return "eMMC";
	default:
		return "UNKNOWN";
	}
}

int rma1_uart_post_setup(int port)
{
	/* setup SCIFA1 Tx/Rx */
	pinmux(PIN_SCIFA_TXD_1, MUX(1, 1, 0, PULL_OFF));
	pinmux(PIN_SCIFA_RXD_1, MUX(1, 0, 1, PULL_UP));

	return 0;
}

static void armadillo8x0_setup_private_data(struct platform_info *pinfo)
{
	struct board_private *priv = pinfo->private_data;
	u8 addr = 0;
	u8 buf[4] = {0, 0, 0, 0};
	struct i2c_device i2c_dev = {
		.port = 1,
		.devid = 0xa0,
		.wait = 5,
	};
	struct i2c_msg msgs[] = {
		[0] = { &addr, 1, I2C_M_WR },
		[1] = { NULL, 6, I2C_M_RD | I2C_REPSTART },
	};

	/* MAC1 Address */
	msgs[1].buf = priv->mac1;
	rma1_i2c_xfer(&i2c_dev, msgs, 2);

	/* USER MAC1 Address */
	addr = 64;
	msgs[1].buf = priv->umac1;
	rma1_i2c_xfer(&i2c_dev, msgs, 2);

	/* Hardware Type */
	addr = 32;
	msgs[1].buf = (u8 *)&priv->hw_type;
	msgs[1].len = 4;
	rma1_i2c_xfer(&i2c_dev, msgs, 2);

	/* Hardware Revision */
	addr = 36;
	msgs[1].buf = (u8 *)&priv->hw_rev;
	msgs[1].len = 2;
	rma1_i2c_xfer(&i2c_dev, msgs, 2);

	/* Lot Number */
	addr = 38;
	msgs[1].buf = (u8 *)&priv->lot_num;
	msgs[1].len = 2;
	rma1_i2c_xfer(&i2c_dev, msgs, 2);

	/* Serial Number */
	addr = 47;
	msgs[1].buf = buf;
	msgs[1].len = 1;
	rma1_i2c_xfer(&i2c_dev, msgs, 2);
	if (buf[0]) {
		addr = 40;
		msgs[1].buf = (u8 *)&priv->serial_num;
		msgs[1].len = 4;
		rma1_i2c_xfer(&i2c_dev, msgs, 2);
	} else {
		priv->serial_num = 0;
	}

	pinfo->system_rev = ((u32)(priv->hw_rev << 16) | (u32)priv->lot_num);
	pinfo->serial_num = priv->serial_num;
	pinfo->machine_no = MACH_ARMADILLO800EVA;

	priv->boot_mode = (read32(SYSC(RESCNT)) >> 24) & 0x7c;
	priv->jumper = gpio_input(xJP1) ? 0 : 1;
	priv->tactsw[0] = gpio_input(xSW1) ? 0 : 1;
	priv->tactsw[1] = gpio_input(xSW2) ? 0 : 1;
	priv->tactsw[2] = gpio_input(xSW3) ? 0 : 1;
	priv->tactsw[3] = gpio_input(xSW4) ? 0 : 1;
}

static void armadillo8x0_get_mac_address(struct platform_info *pinfo, u8 *addr)
{
	struct board_private *priv = pinfo->private_data;
	u8 *mac;

	if (is_valid_mac(priv->umac1))
		mac = priv->umac1;
	else
		mac = priv->mac1;

	memcpy(addr, mac, 6);
}

static void armadillo8x0_udelay(struct platform_info *pinfo, u32 usecs)
{
	u32 count;

	/* SUB = 50MHz, TCNT_CLK = 12.5MHz */
	count = (12 * usecs) + (usecs / 2) + (usecs & 1);

	write16(TMU(0, TCR0), 0);
	write32(TMU(0, TCOR0), count);
	write32(TMU(0, TCNT0), count);

	write8(TMU(0, TSTR), 1);
	while (!(read16(TMU(0, TCR0)) & 0x100));
	write16(TMU(0, TCR0), 0x0000);
	write8(TMU(0, TSTR), 0);
}

static void armadillo8x0_led_on(struct platform_info *pinfo, u32 leds)
{
	if (leds & LED_RED)
		gpio_output(xLED1, 0);
	if (leds & LED_GREEN)
		gpio_output(xLED2, 0);
}

static void armadillo8x0_led_off(struct platform_info *pinfo, u32 leds)
{
	if (leds & LED_RED)
		gpio_output(xLED1, 1);
	if (leds & LED_GREEN)
		gpio_output(xLED2, 1);
}

static struct page_table pt_list[] = {
	/* Physical,  Virtual,    Size,       Option */
	/*-------------------------------------------*/
	/* SDRAM */
	{0x40000000, 0x40000000, 0x00b00000, 0xc0e},
	{0x40b00000, 0x40b00000, 0x00100000, 0xc02},
	{0x40c00000, 0x40c00000, 0x1f400000, 0xc0e},
	/* Peripheral modules */
	{0xe6000000, 0xe6000000, 0x02000000, 0xc02},
	{0xffc00000, 0xffc00000, 0x00400000, 0xc02},
	/* GbEther */
	{0xe9a00000, 0xe9a00000, 0x00200000, 0xc02},
};

static int armadillo8x0_get_mmu_page_tables(struct platform_info *pinfo,
					    u32 mode, struct page_table **pt,
					    int *nr_pt)
{
	*pt = (struct page_table *)&pt_list;
	*nr_pt = ARRAY_SIZE(pt_list);

	return 0;
}

#define on_off_string(x) ((x) ? "ON" : "OFF")
static void armadillo8x0_display_info(struct platform_info *pinfo)
{
	struct board_private *priv = pinfo->private_data;

	hprintf("   Board Type: %p\n"
		"     Revision: %p\n"
		"          Lot: %p\n"
		"Serial Number: %d\n"
		"    Boot Mode: %p (%s)\n"
		"       Jumper: %s\n"
		"      Tact-SW: %s,%s,%s,%s\n",
		priv->hw_type, priv->hw_rev, priv->lot_num, priv->serial_num, 
		priv->boot_mode,
		get_boot_mode_str(priv->boot_mode),
		on_off_string(priv->jumper),
		on_off_string(priv->tactsw[0]),
		on_off_string(priv->tactsw[1]),
		on_off_string(priv->tactsw[2]),
		on_off_string(priv->tactsw[3]));

	hprintf(" ORIG MAC-1: ");
	print_mac(priv->mac1);
	if (is_valid_mac(priv->umac1)) {
		hprintf(" USER MAC-1: ");
		print_mac(priv->umac1);
	}
}

static void adjust_recovery_boot_args(struct platform_info *pinfo,
				      int *argc, char *argv[])
{
	*argc = 5;
	argv[1] = "console=ttySC1,115200";
	argv[2] = "noinitrd";
	argv[3] = "rootwait";
	argv[4] = "root=/dev/mmcblk1p2";
}

static void adjust_recovery_boot_device(struct platform_info *pinfo,
					char *device)
{
	hsnprintf(device, CMDLINE_MAXLEN, "mmcblk1p2");
}

static int armadillo8x0_is_autoboot(struct platform_info *pinfo)
{
	struct board_private *priv = pinfo->private_data;
	int autoboot = priv->jumper ? 0 : 1;

	if (priv->boot_mode == BOOT_MODE_SD &&
	    priv->tactsw[0]) {
		hprintf("*** Recovery Mode ******************************\n");
		autoboot = 1;
	    	pinfo->adjust_boot_args = adjust_recovery_boot_args;
	    	pinfo->adjust_boot_device = adjust_recovery_boot_device;
	}

	return autoboot;
}

static void armadillo8x0_prepare_normal_boot(struct platform_info *pinfo)
{
	pinfo->adjust_boot_args = NULL;
	pinfo->adjust_boot_device = NULL;
}

static void armadillo8x0_setup_map(struct platform_info *pinfo)
{
	size_t size;

	size = DRAM_SIZE;
	pinfo->map->ram.size = size;
	hsprintf(pinfo->default_ramparts, "%p(dram-1)", size);
#if !defined(CONFIG_SYSTEM_ENTRYPOINT_FLASH)
	pinfo->map->flash.size = 16*1024*1024;
	hsprintf(pinfo->default_mtdparts,
		 "armadillo8x0-nor:%p(all)ro,0x20000@0(bootloader)ro,"
		 "0x600000(kernel),%p(userland),-(config)",
		 pinfo->map->flash.size, pinfo->map->flash.size - 0x640000);
#endif
}

static void update_target_profile(struct platform_info *pinfo)
{
	char *ptr = target_name;

	ptr += hsprintf(ptr, "%s", "Armadillo-800 EVA");
	if (strlen(CONFIG_PROFILE_NAME))
		hsprintf(ptr, "/%s", CONFIG_PROFILE_NAME);
}

static void armadillo8x0_setup_console(struct platform_info *pinfo)
{
	char console[128];
	int ret;
	int autoboot = armadillo8x0_is_autoboot(pinfo);

	hsprintf(console, "%s", CONFIG_DEFAULT_CONSOLE);

	/* we must set up to STANDARD_CONSOLE to match our hardware.
	   FIXME: this is redundant and just using a side effect of
	   the function to initialize it */
	change_console(CONFIG_STANDARD_CONSOLE);
	ret = change_console(console);
	if (ret && !autoboot)
		change_console(CONFIG_STANDARD_CONSOLE);
}

static void armadillo8x0_setup_ethernet(struct platform_info *pinfo)
{
	struct net_device *ndev = pinfo->net_dev;

	cpg_clr_mstpcr(309);
	if (ndev && ndev->set_mac) {
		armadillo8x0_get_mac_address(pinfo, ndev->eth_mac);
		ndev->set_mac(ndev, ndev->eth_mac);
	}
}

static void armadillo8x0_init(void)
{
	struct platform_info *pinfo = &platform_info;

	/* disable cache on ICB */
	write32(0xfe950098, 0x01600164);

	/* Module Enable */
	cpg_clr_mstpcr(125); /* TMU0 */

	/* Setup Pinmux */
	pinmux(xJP1,		MUX(0, 0, 1, PULL_OFF));
	pinmux(xSW1,		MUX(0, 0, 1, PULL_OFF));
	pinmux(xSW2,		MUX(0, 0, 1, PULL_OFF));
	pinmux(xSW3,		MUX(0, 0, 1, PULL_OFF));
	pinmux(xSW4,		MUX(0, 0, 1, PULL_OFF));
	pinmux(xLED1,		MUX(0, 1, 0, PULL_OFF));
	pinmux(xLED2,		MUX(0, 1, 0, PULL_OFF));
	pinmux(xLED3,		MUX(0, 1, 0, PULL_OFF));
	pinmux(xLED4,		MUX(0, 1, 0, PULL_OFF));

	/* LED2/3/4 OFF */
	gpio_output(xLED2, 0);
	gpio_output(xLED3, 0);
	gpio_output(xLED4, 0);

	rma1_setup_i2c(1);
	armadillo8x0_setup_private_data(pinfo);
	update_target_profile(pinfo);

	memset((void *)RAM_ADDR(0x05000000), 0, 0x00020000);
	armadillo8x0_setup_map(pinfo);
	armadillo8x0_setup_console(pinfo);
	armadillo8x0_setup_ethernet(pinfo);

	memdev_add(emmcboot1, &mdev_param,
		   "hermit/param", 0x00000000, 0x00002000);
}
arch_initcall(armadillo8x0_init);

static struct memory_map armadillo8x0_memory_map = {
	/* default memory map: RAM = 64MB, (FLA = 16MB) */
	.ram		= { RAM_ADDR(0x00000000), 0 },
	.boot_param	= { RAM_ADDR(0x00000100), 0x00000f00 },
	.mmu_table	= { RAM_ADDR(0x00004000), 0x00004000 },
	.kernel		= { RAM_ADDR(0x00008000), 0x007f8000 },
/*	.hermit		= { RAM_ADDR(0x00800000), 0x00300000 }, */
/*	.fec_desc	= { RAM_ADDR(0x00b00000), 0x00100000 }, */
	.gunzip		= { RAM_ADDR(0x00c00000), 0x00100000 },
/*	.stack(irq)	= { RAM_ADDR(0x00d00000), 0x00100000 }, */
/*	.stack(svc)	= { RAM_ADDR(0x00e00000), 0x00100000 }, */
/*	.vector		= { RAM_ADDR(0x00f00000), 0x00100000 }, */
	.initrd		= { RAM_ADDR(0x01000000), 0x02000000 },
	.free		= { RAM_ADDR(0x03000000), 0x01000000 },
	.param          = { RAM_ADDR(0x05018000), 0x00008000 },
};

struct boot_device boot_device_table[] = {
	/*			minor	minor	*/
	/* name		major	start	inc	*/
	{"ram",		1,	0,	1,	},
	{"mtdblock",	31,	0,	1,	},
	{"mmcblk0p",	0,	1,	1,	},
};

struct platform_info platform_info = {
	.prepare_normal_boot = armadillo8x0_prepare_normal_boot,
	.is_autoboot = armadillo8x0_is_autoboot,
	.led_on = armadillo8x0_led_on,
	.led_off = armadillo8x0_led_off,
	.udelay = armadillo8x0_udelay,
	.get_mmu_page_tables = armadillo8x0_get_mmu_page_tables,
	.display_info = armadillo8x0_display_info,
	.get_mac_address = armadillo8x0_get_mac_address,

	.private_data = &board_priv,
	.map = &armadillo8x0_memory_map,

	.net_dev = &rma1_eth,
	.mmcsd_dev = &rma1_mmcsd,
};
